"""fontTools.feaLib -- a package for dealing with OpenType feature files."""

# The structure of OpenType feature files is defined here:
# http://www.adobe.com/devnet/opentype/afdko/topic_feature_file_syntax.html



from fontTools import agl
import unittest


class AglToUnicodeTest(unittest.TestCase):
    def test_spec_examples(self):
        # https://github.com/adobe-type-tools/agl-specification#3-examples
        self.assertEqual(agl.toUnicode("Lcommaaccent"), "Ļ")
        self.assertEqual(agl.toUnicode("uni20AC0308"), "\u20AC\u0308")
        self.assertEqual(agl.toUnicode("u1040C"), "\U0001040C")
        self.assertEqual(agl.toUnicode("uniD801DC0C"), "")
        self.assertEqual(agl.toUnicode("uni20ac"), "")
        self.assertEqual(
            agl.toUnicode("Lcommaaccent_uni20AC0308_u1040C.alternate"),
            "\u013B\u20AC\u0308\U0001040C",
        )
        self.assertEqual(agl.toUnicode("Lcommaaccent_uni013B_u013B"), "ĻĻĻ")
        self.assertEqual(agl.toUnicode("foo"), "")
        self.assertEqual(agl.toUnicode(".notdef"), "")

    def test_aglfn(self):
        self.assertEqual(agl.toUnicode("longs_t"), "ſt")
        self.assertEqual(agl.toUnicode("f_f_i.alt123"), "ffi")

    def test_uniABCD(self):
        self.assertEqual(agl.toUnicode("uni0041"), "A")
        self.assertEqual(agl.toUnicode("uni0041_uni0042_uni0043"), "ABC")
        self.assertEqual(agl.toUnicode("uni004100420043"), "ABC")
        self.assertEqual(agl.toUnicode("uni"), "")
        self.assertEqual(agl.toUnicode("uni41"), "")
        self.assertEqual(agl.toUnicode("uni004101"), "")
        self.assertEqual(agl.toUnicode("uniDC00"), "")

    def test_uABCD(self):
        self.assertEqual(agl.toUnicode("u0041"), "A")
        self.assertEqual(agl.toUnicode("u00041"), "A")
        self.assertEqual(agl.toUnicode("u000041"), "A")
        self.assertEqual(agl.toUnicode("u0000041"), "")
        self.assertEqual(agl.toUnicode("u0041_uni0041_A.alt"), "AAA")

    def test_union(self):
        # Interesting test case because "uni" is a prefix of "union".
        self.assertEqual(agl.toUnicode("union"), "∪")
        # U+222A U+FE00 is a Standardized Variant for UNION WITH SERIFS.
        self.assertEqual(agl.toUnicode("union_uniFE00"), "\u222A\uFE00")

    def test_dingbats(self):
        self.assertEqual(agl.toUnicode("a20", isZapfDingbats=True), "✔")
        self.assertEqual(agl.toUnicode("a20.alt", isZapfDingbats=True), "✔")
        self.assertEqual(agl.toUnicode("a206", isZapfDingbats=True), "❰")
        self.assertEqual(agl.toUnicode("a20", isZapfDingbats=False), "")
        self.assertEqual(agl.toUnicode("a0", isZapfDingbats=True), "")
        self.assertEqual(agl.toUnicode("a207", isZapfDingbats=True), "")
        self.assertEqual(agl.toUnicode("abcdef", isZapfDingbats=True), "")


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



#! /usr/bin/env python3

"""
Sample script to use the otlLib.optimize.gpos functions to compact GPOS tables
of existing fonts. This script takes one or more TTF files as arguments and
will create compacted copies of the fonts using all available modes of the GPOS
compaction algorithm. For each copy, it will measure the new size of the GPOS
table and also the new size of the font in WOFF2 format. All results will be
printed to stdout in CSV format, so the savings provided by the algorithm in
each mode can be inspected.

This was initially made to debug the algorithm but can also be used to choose
a mode value for a specific font (trade-off between bytes saved in TTF format
vs more bytes in WOFF2 format and more subtables).

Run:

python Snippets/compact_gpos.py MyFont.ttf > results.csv
"""

import argparse
from collections import defaultdict
import csv
import time
import sys
from pathlib import Path
from typing import Any, Iterable, List, Optional, Sequence, Tuple

from fontTools.ttLib import TTFont
from fontTools.otlLib.optimize import compact

MODES = [str(c) for c in range(1, 10)]


def main(args: Optional[List[str]] = None):
    parser = argparse.ArgumentParser()
    parser.add_argument("fonts", type=Path, nargs="+", help="Path to TTFs.")
    parsed_args = parser.parse_args(args)

    runtimes = defaultdict(list)
    rows = []
    font_path: Path
    for font_path in parsed_args.fonts:
        font = TTFont(font_path)
        if "GPOS" not in font:
            print(f"No GPOS in {font_path.name}, skipping.", file=sys.stderr)
            continue
        size_orig = len(font.getTableData("GPOS")) / 1024
        print(f"Measuring {font_path.name}...", file=sys.stderr)

        fonts = {}
        font_paths = {}
        sizes = {}
        for mode in MODES:
            print(f"    Running mode={mode}", file=sys.stderr)
            fonts[mode] = TTFont(font_path)
            before = time.perf_counter()
            compact(fonts[mode], mode=str(mode))
            runtimes[mode].append(time.perf_counter() - before)
            font_paths[mode] = (
                font_path.parent
                / "compact"
                / (font_path.stem + f"_{mode}" + font_path.suffix)
            )
            font_paths[mode].parent.mkdir(parents=True, exist_ok=True)
            fonts[mode].save(font_paths[mode])
            fonts[mode] = TTFont(font_paths[mode])
            sizes[mode] = len(fonts[mode].getTableData("GPOS")) / 1024

        print(f"    Runtimes:", file=sys.stderr)
        for mode, times in runtimes.items():
            print(
                f"        {mode:10} {' '.join(f'{t:5.2f}' for t in times)}",
                file=sys.stderr,
            )

        # Bonus: measure WOFF2 file sizes.
        print(f"    Measuring WOFF2 sizes", file=sys.stderr)
        size_woff_orig = woff_size(font, font_path) / 1024
        sizes_woff = {
            mode: woff_size(fonts[mode], font_paths[mode]) / 1024 for mode in MODES
        }

        rows.append(
            (
                font_path.name,
                size_orig,
                size_woff_orig,
                *flatten(
                    (
                        sizes[mode],
                        pct(sizes[mode], size_orig),
                        sizes_woff[mode],
                        pct(sizes_woff[mode], size_woff_orig),
                    )
                    for mode in MODES
                ),
            )
        )

    write_csv(rows)


def woff_size(font: TTFont, path: Path) -> int:
    font.flavor = "woff2"
    woff_path = path.with_suffix(".woff2")
    font.save(woff_path)
    return woff_path.stat().st_size


def write_csv(rows: List[Tuple[Any]]) -> None:
    sys.stdout.reconfigure(encoding="utf-8")
    sys.stdout.write("\uFEFF")
    writer = csv.writer(sys.stdout, lineterminator="\n")
    writer.writerow(
        [
            "File",
            "Original GPOS Size",
            "Original WOFF2 Size",
            *flatten(
                (
                    f"mode={mode}",
                    f"Change {mode}",
                    f"mode={mode} WOFF2 Size",
                    f"Change {mode} WOFF2 Size",
                )
                for mode in MODES
            ),
        ]
    )
    for row in rows:
        writer.writerow(row)


def pct(new: float, old: float) -> float:
    return -(1 - (new / old))


def flatten(seq_seq: Iterable[Iterable[Any]]) -> List[Any]:
    return [thing for seq in seq_seq for thing in seq]


if __name__ == "__main__":
    main()



import logging
import shutil

import pytest

from fontTools.misc import plistlib
from fontTools.ufoLib import UFOFormatVersion, UFOReader, UFOWriter
from fontTools.ufoLib.errors import UFOLibError, UnsupportedUFOFormat


@pytest.fixture
def ufo_path(tmp_path):
    ufodir = tmp_path / "TestFont.ufo"
    ufodir.mkdir()
    with (ufodir / "metainfo.plist").open("wb") as f:
        plistlib.dump({"creator": "pytest", "formatVersion": 3}, f)
    (ufodir / "glyphs").mkdir()
    with (ufodir / "layercontents.plist").open("wb") as f:
        plistlib.dump([("public.default", "glyphs")], f)
    return ufodir


def test_formatVersion_deprecated(ufo_path):
    reader = UFOReader(ufo_path)

    with pytest.warns(DeprecationWarning) as warnings:
        assert reader.formatVersion == 3

    assert len(warnings) == 1
    assert "is deprecated; use the 'formatVersionTuple'" in warnings[0].message.args[0]


def test_formatVersionTuple(ufo_path):
    reader = UFOReader(ufo_path)

    assert reader.formatVersionTuple == (3, 0)
    assert reader.formatVersionTuple.major == 3
    assert reader.formatVersionTuple.minor == 0
    assert str(reader.formatVersionTuple) == "3.0"


def test_readMetaInfo_errors(ufo_path):
    (ufo_path / "metainfo.plist").unlink()
    with pytest.raises(UFOLibError, match="'metainfo.plist' is missing"):
        UFOReader(ufo_path)

    (ufo_path / "metainfo.plist").write_bytes(plistlib.dumps({}))
    with pytest.raises(UFOLibError, match="Missing required formatVersion"):
        UFOReader(ufo_path)

    (ufo_path / "metainfo.plist").write_bytes(plistlib.dumps([]))
    with pytest.raises(UFOLibError, match="metainfo.plist is not properly formatted"):
        UFOReader(ufo_path)


def test_readMetaInfo_unsupported_format_version(ufo_path, caplog):
    metainfo = {"formatVersion": 10, "formatVersionMinor": 15}
    (ufo_path / "metainfo.plist").write_bytes(plistlib.dumps(metainfo))

    with pytest.raises(UnsupportedUFOFormat):
        UFOReader(ufo_path)  # validate=True by default

    with pytest.raises(UnsupportedUFOFormat):
        UFOReader(ufo_path, validate=True)

    caplog.clear()
    with caplog.at_level(logging.WARNING, logger="fontTools.ufoLib"):
        UFOReader(ufo_path, validate=False)

    assert len(caplog.records) == 1
    assert "Unsupported UFO format" in caplog.text
    assert "Assuming the latest supported version" in caplog.text


def test_UFOWriter_formatVersion(tmp_path):
    ufo_path = tmp_path / "TestFont.ufo"
    with UFOWriter(ufo_path, formatVersion=3) as writer:
        assert writer.formatVersionTuple == (3, 0)

    shutil.rmtree(str(ufo_path))
    with UFOWriter(ufo_path, formatVersion=(2, 0)) as writer:
        assert writer.formatVersionTuple == (2, 0)


def test_UFOWriter_formatVersion_default_latest(tmp_path):
    writer = UFOWriter(tmp_path / "TestFont.ufo")
    assert writer.formatVersionTuple == UFOFormatVersion.default()


def test_UFOWriter_unsupported_format_version(tmp_path):
    with pytest.raises(UnsupportedUFOFormat):
        UFOWriter(tmp_path, formatVersion=(123, 456))


def test_UFOWriter_previous_higher_format_version(ufo_path):
    with pytest.raises(
        UnsupportedUFOFormat, match="UFO located at this path is a higher version"
    ):
        UFOWriter(ufo_path, formatVersion=(2, 0))


def test_getFileModificationTime(ufo_path):
    import time

    with UFOWriter(ufo_path) as writer:
        raw_metainfo = writer.readBytesFromPath("metainfo.plist")
        modified = writer.getFileModificationTime("metainfo.plist")
        assert isinstance(modified, float)
        # writeBytesToPath only modifies a file if the content to be written is not the same
        writer.writeBytesToPath("metainfo.plist", raw_metainfo)
        assert writer.getFileModificationTime("metainfo.plist") == modified
        time.sleep(0.1)
        writer.writeBytesToPath("metainfo.plist", raw_metainfo + b"\n")
        assert writer.getFileModificationTime("metainfo.plist") != modified



# Copyright 2023 Behdad Esfahbod. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import pytest

from fontTools.qu2cu import quadratic_to_curves
from fontTools.qu2cu.qu2cu import main as qu2cu_main
from fontTools.qu2cu.benchmark import main as benchmark_main

import os
import json
from fontTools.cu2qu import curve_to_quadratic


class Qu2CuTest:
    def test_empty_input(self):
        assert quadratic_to_curves([]) == []

    @pytest.mark.parametrize(
        "quadratics, expected, tolerance, cubic_only",
        [
            (
                [
                    [(0, 0), (0, 1), (2, 1), (2, 0)],
                ],
                [
                    ((0, 0), (0, 4 / 3), (2, 4 / 3), (2, 0)),
                ],
                0.1,
                True,
            ),
            (
                [
                    [(0, 0), (0, 1), (2, 1), (2, 2)],
                ],
                [
                    ((0, 0), (0, 4 / 3), (2, 2 / 3), (2, 2)),
                ],
                0.2,
                True,
            ),
            (
                [
                    [(0, 0), (0, 1), (1, 1)],
                    [(1, 1), (3, 1), (3, 0)],
                ],
                [
                    ((0, 0), (0, 1), (1, 1)),
                    ((1, 1), (3, 1), (3, 0)),
                ],
                0.2,
                False,
            ),
            (
                [
                    [(0, 0), (0, 1), (1, 1)],
                    [(1, 1), (3, 1), (3, 0)],
                ],
                [
                    ((0, 0), (0, 2 / 3), (1 / 3, 1), (1, 1)),
                    ((1, 1), (7 / 3, 1), (3, 2 / 3), (3, 0)),
                ],
                0.2,
                True,
            ),
        ],
    )
    def test_simple(self, quadratics, expected, tolerance, cubic_only):
        expected = [
            tuple((pytest.approx(p[0]), pytest.approx(p[1])) for p in curve)
            for curve in expected
        ]

        c = quadratic_to_curves(quadratics, tolerance, cubic_only)
        assert c == expected

    def test_roundtrip(self):
        DATADIR = os.path.join(os.path.dirname(__file__), "..", "cu2qu", "data")
        with open(os.path.join(DATADIR, "curves.json"), "r") as fp:
            curves = json.load(fp)

        tolerance = 1

        splines = [curve_to_quadratic(c, tolerance) for c in curves]
        reconsts = [quadratic_to_curves([spline], tolerance) for spline in splines]

        for curve, reconst in zip(curves, reconsts):
            assert len(reconst) == 1
            curve = tuple((pytest.approx(p[0]), pytest.approx(p[1])) for p in curve)
            assert curve == reconst[0]

    def test_reject_disconnected_splines(self):
        with pytest.raises(ValueError, match="must connect end-to-start"):
            quadratic_to_curves(
                [
                    [(0, 0), (0, 1), (1, 1)],
                    [(2, 2), (3, 3), (4, 4)],
                ]
            )

    def test_reject_short_spline(self):
        with pytest.raises(ValueError, match="at least 3 points"):
            quadratic_to_curves([[(0, 0), (1, 1)]])

    @pytest.mark.parametrize("tolerance", [0, -1])
    def test_reject_non_positive_tolerance(self, tolerance):
        with pytest.raises(ValueError, match="max_err must be greater than zero"):
            quadratic_to_curves([[(0, 0), (0, 1), (1, 1)]], tolerance)

    def test_main(self):
        # Just for coverage
        qu2cu_main()
        benchmark_main()



from fontTools.misc.configTools import AbstractConfig, Options
import pytest
from fontTools.config import (
    OPTIONS,
    Config,
    ConfigUnknownOptionError,
    ConfigValueParsingError,
    ConfigValueValidationError,
)
from fontTools.ttLib import TTFont


def test_can_register_option():
    MY_OPTION = Config.register_option(
        name="tests:MY_OPTION",
        help="Test option, value should be True or False, default = True",
        default=True,
        parse=lambda v: v in ("True", "true", 1, True),
        validate=lambda v: v == True or v == False,
    )

    assert MY_OPTION.name == "tests:MY_OPTION"
    assert (
        MY_OPTION.help == "Test option, value should be True or False, default = True"
    )
    assert MY_OPTION.default == True
    assert MY_OPTION.parse("true") == True
    assert MY_OPTION.validate("hello") == False

    ttFont = TTFont(cfg={"tests:MY_OPTION": True})
    assert True == ttFont.cfg.get("tests:MY_OPTION")
    assert True == ttFont.cfg.get(MY_OPTION)


# to parametrize tests of Config mapping interface accepting either a str or Option
@pytest.fixture(
    params=[
        pytest.param("fontTools.otlLib.optimize.gpos:COMPRESSION_LEVEL", id="str"),
        pytest.param(
            OPTIONS["fontTools.otlLib.optimize.gpos:COMPRESSION_LEVEL"], id="Option"
        ),
    ]
)
def COMPRESSION_LEVEL(request):
    return request.param


def test_ttfont_has_config(COMPRESSION_LEVEL):
    ttFont = TTFont(cfg={COMPRESSION_LEVEL: 8})
    assert 8 == ttFont.cfg.get(COMPRESSION_LEVEL)


def test_ttfont_can_take_superset_of_fonttools_config(COMPRESSION_LEVEL):
    # Create MyConfig with all options from fontTools.config plus some
    my_options = Options(Config.options)
    MY_OPTION = my_options.register("custom:my_option", "help", "default", str, any)

    class MyConfig(AbstractConfig):
        options = my_options

    ttFont = TTFont(cfg=MyConfig({"custom:my_option": "my_value"}))
    assert 0 == ttFont.cfg.get(COMPRESSION_LEVEL)
    assert "my_value" == ttFont.cfg.get(MY_OPTION)

    # but the default Config doens't know about MY_OPTION
    with pytest.raises(ConfigUnknownOptionError):
        TTFont(cfg={MY_OPTION: "my_value"})


def test_no_config_returns_default_values(COMPRESSION_LEVEL):
    ttFont = TTFont()
    assert 0 == ttFont.cfg.get(COMPRESSION_LEVEL)
    assert 3 == ttFont.cfg.get(COMPRESSION_LEVEL, 3)


def test_can_set_config(COMPRESSION_LEVEL):
    ttFont = TTFont()
    ttFont.cfg.set(COMPRESSION_LEVEL, 5)
    assert 5 == ttFont.cfg.get(COMPRESSION_LEVEL)
    ttFont.cfg.set(COMPRESSION_LEVEL, 6)
    assert 6 == ttFont.cfg.get(COMPRESSION_LEVEL)


def test_different_ttfonts_have_different_configs(COMPRESSION_LEVEL):
    cfg = Config({COMPRESSION_LEVEL: 5})
    ttFont1 = TTFont(cfg=cfg)
    ttFont2 = TTFont(cfg=cfg)
    ttFont2.cfg.set(COMPRESSION_LEVEL, 6)
    assert 5 == ttFont1.cfg.get(COMPRESSION_LEVEL)
    assert 6 == ttFont2.cfg.get(COMPRESSION_LEVEL)


def test_cannot_set_inexistent_key():
    with pytest.raises(ConfigUnknownOptionError):
        TTFont(cfg={"notALib.notAModule.inexistent": 4})


def test_value_not_parsed_by_default(COMPRESSION_LEVEL):
    # Note: value given as a string
    with pytest.raises(ConfigValueValidationError):
        TTFont(cfg={COMPRESSION_LEVEL: "8"})


def test_value_gets_parsed_if_asked(COMPRESSION_LEVEL):
    # Note: value given as a string
    ttFont = TTFont(cfg=Config({COMPRESSION_LEVEL: "8"}, parse_values=True))
    assert 8 == ttFont.cfg.get(COMPRESSION_LEVEL)


def test_value_parsing_can_error(COMPRESSION_LEVEL):
    with pytest.raises(ConfigValueParsingError):
        TTFont(
            cfg=Config(
                {COMPRESSION_LEVEL: "not an int"},
                parse_values=True,
            )
        )


def test_value_gets_validated(COMPRESSION_LEVEL):
    # Note: 12 is not a valid value for GPOS compression level (must be in 0-9)
    with pytest.raises(ConfigValueValidationError):
        TTFont(cfg={COMPRESSION_LEVEL: 12})


def test_implements_mutable_mapping(COMPRESSION_LEVEL):
    cfg = Config()
    cfg[COMPRESSION_LEVEL] = 2
    assert 2 == cfg[COMPRESSION_LEVEL]
    assert list(iter(cfg))
    assert 1 == len(cfg)
    del cfg[COMPRESSION_LEVEL]
    assert 0 == cfg[COMPRESSION_LEVEL]
    assert not list(iter(cfg))
    assert 0 == len(cfg)
